@@ -48,9 +48,10 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 	int is_26 = get_linux_version() >= 2006000;
 	int is_rebuilding = 0;
 	int failed = 0;
-	struct supertype *st = NULL;
+	struct supertype *st;
+	char *subarray = NULL;
 	int max_disks = MD_SB_DISKS; /* just a default */
-	struct mdinfo info;
+	struct mdinfo *info = NULL;
 	struct mdinfo *sra;
 	char *member = NULL;
 	char *container = NULL;
@@ -88,7 +89,7 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 		return rv;
 	}
 	sra = sysfs_read(fd, 0, GET_VERSION);
-	st = super_by_fd(fd);
+	st = super_by_fd(fd, &subarray);
 
 	if (fstat(fd, &stb) != 0 && !S_ISBLK(stb.st_mode))
 		stb.st_rdev = 0;
@@ -97,23 +98,22 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 	if (st)
 		max_disks = st->max_devs;
 
-	if (sra && is_subarray(sra->text_version) &&
-		strchr(sra->text_version+1, '/')) {
+	if (subarray) {
 		/* This is a subarray of some container.
 		 * We want the name of the container, and the member
 		 */
-		char *s = strchr(sra->text_version+1, '/');
-		int dn;
-		*s++ = '\0';
-		member = s;
-		dn = devname2devnum(sra->text_version+1);
+		int dn = st->container_dev;
+
+		member = subarray;
 		container = map_dev(dev2major(dn), dev2minor(dn), 1);
 	}
 
 	/* try to load a superblock */
-	for (d= 0; d<max_disks; d++) {
+	if (st) for (d = 0; d < max_disks; d++) {
 		mdu_disk_info_t disk;
 		char *dv;
+		int fd2;
+		int err;
 		disk.number = d;
 		if (ioctl(fd, GET_DISK_INFO, &disk) < 0)
 			continue;
@@ -121,38 +121,58 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 		    disk.major == 0 &&
 		    disk.minor == 0)
 			continue;
-		if ((dv=map_dev(disk.major, disk.minor, 1))) {
-			/* some formats (imsm) have free-floating-spares
-			 * with a uuid of uuid_match_any, they don't
-			 * have very good info about the rest of the
-			 * container, so keep searching when
-			 * encountering such a device.  Otherwise, stop
-			 * after the first successful call to
-			 * ->load_super.
-			 */
-			int free_spare = memcmp(uuid_match_any,
-						info.uuid,
-						sizeof(uuid_match_any)) == 0;
-			if ((!st || !st->sb || free_spare) &&
-			    (array.raid_disks == 0 || 
-			     (disk.state & (1<<MD_DISK_ACTIVE)))) {
-				/* try to read the superblock from this device
-				 * to get more info
-				 */
-				if (free_spare)
-					st->ss->free_super(st);
-				int fd2 = dev_open(dv, O_RDONLY);
-				if (fd2 >=0 && st &&
-				    st->ss->load_super(st, fd2, NULL) == 0) {
-					st->ss->getinfo_super(st, &info);
-					if (array.raid_disks != 0 && /* container */
-					    (info.array.ctime != array.ctime ||
-					     info.array.level != array.level))
-						st->ss->free_super(st);
-				}
-				if (fd2 >= 0) close(fd2);
-			}
+
+		if (array.raid_disks > 0 &&
+		    (disk.state & (1 << MD_DISK_ACTIVE)) == 0)
+			continue;
+
+		dv = map_dev(disk.major, disk.minor, 1);
+		if (!dv)
+			continue;
+
+		fd2 = dev_open(dv, O_RDONLY);
+		if (fd2 < 0)
+			continue;
+
+		if (st->sb)
+			st->ss->free_super(st);
+
+		err = st->ss->load_super(st, fd2, NULL);
+		close(fd2);
+		if (err)
+			continue;
+		if (info)
+			free(info);
+		if (subarray)
+			info = st->ss->container_content(st, subarray);
+		else {
+			info = malloc(sizeof(*info));
+			st->ss->getinfo_super(st, info, NULL);
 		}
+		if (!info)
+			continue;
+
+		if (array.raid_disks != 0 && /* container */
+		    (info->array.ctime != array.ctime ||
+		     info->array.level != array.level)) {
+			st->ss->free_super(st);
+			continue;
+		}
+		/* some formats (imsm) have free-floating-spares
+		 * with a uuid of uuid_zero, they don't
+		 * have very good info about the rest of the
+		 * container, so keep searching when
+		 * encountering such a device.  Otherwise, stop
+		 * after the first successful call to
+		 * ->load_super.
+		 */
+		if (memcmp(uuid_zero,
+			   info->uuid,
+			   sizeof(uuid_zero)) == 0) {
+			st->ss->free_super(st);
+			continue;
+		}
+		break;
 	}
 
 	/* Ok, we have some info to print... */
@@ -178,14 +198,13 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 				       array.major_version, array.minor_version);
 		}
 		
-		if (st && st->sb) {
-			struct mdinfo info;
+		if (st && st->sb && info) {
 			char nbuf[64];
 			struct map_ent *mp, *map = NULL;
-			st->ss->getinfo_super(st, &info);
-			fname_from_uuid(st, &info, nbuf, ':');
+
+			fname_from_uuid(st, info, nbuf, ':');
 			printf("MD_UUID=%s\n", nbuf+5);
-			mp = map_by_uuid(&map, info.uuid);
+			mp = map_by_uuid(&map, info->uuid);
 			if (mp && mp->path &&
 			    strncmp(mp->path, "/dev/md/", 8) == 0)
 				printf("MD_DEVNAME=%s\n", mp->path+8);
@@ -278,7 +297,7 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 		struct mdstat_ent *ms = mdstat_read(0, 0);
 		struct mdstat_ent *e;
 		int devnum = array.md_minor;
-		if (major(stb.st_rdev) == get_mdp_major())
+		if (major(stb.st_rdev) == (unsigned)get_mdp_major())
 			devnum = -1 - devnum;
 
 		for (e=ms; e; e=e->next)
@@ -343,6 +362,7 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 		if (atime)
 			printf("    Update Time : %.24s\n", ctime(&atime));
 		if (array.raid_disks) {
+			static char *sync_action[] = {", recovering",", resyncing",", reshaping",", checking"};
 			char *st;
 			if (avail_disks == array.raid_disks)
 				st = "";
@@ -355,8 +375,7 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 			printf("          State : %s%s%s%s\n",
 			       (array.state&(1<<MD_SB_CLEAN))?"clean":"active",
 			       st,
-			       (!e || e->percent < 0) ? "" :
-			       (e->resync) ? ", resyncing": ", recovering",
+			       (!e || e->percent < 0) ? "" : sync_action[e->resync],
 			       larray_size ? "": ", Not Started");
 		}
 		if (array.raid_disks)
@@ -398,50 +417,50 @@ int Detail(char *dev, int brief, int export, int test, char *homehost)
 
 		if (e && e->percent >= 0) {
 			printf(" Re%s Status : %d%% complete\n",
-			       (st && st->sb && info.reshape_active)?
+			       (st && st->sb && info->reshape_active)?
 			          "shape":"build",
 			       e->percent);
 			is_rebuilding = 1;
 		}
 		free_mdstat(ms);
 
-		if (st->sb && info.reshape_active) {
+		if (st->sb && info->reshape_active) {
 #if 0
 This is pretty boring
-			printf("  Reshape pos'n : %llu%s\n", (unsigned long long) info.reshape_progress<<9,
-			       human_size(info.reshape_progress<<9));
+			printf("  Reshape pos'n : %llu%s\n", (unsigned long long) info->reshape_progress<<9,
+			       human_size((unsigned long long)info->reshape_progress<<9));
 #endif
-			if (info.delta_disks > 0)
+			if (info->delta_disks > 0)
 				printf("  Delta Devices : %d, (%d->%d)\n",
-				       info.delta_disks, array.raid_disks - info.delta_disks, array.raid_disks);
-			if (info.delta_disks < 0)
+				       info->delta_disks, array.raid_disks - info->delta_disks, array.raid_disks);
+			if (info->delta_disks < 0)
 				printf("  Delta Devices : %d, (%d->%d)\n",
-				       info.delta_disks, array.raid_disks, array.raid_disks + info.delta_disks);
-			if (info.new_level != array.level) {
-				char *c = map_num(pers, info.new_level);
+				       info->delta_disks, array.raid_disks, array.raid_disks + info->delta_disks);
+			if (info->new_level != array.level) {
+				char *c = map_num(pers, info->new_level);
 				printf("      New Level : %s\n", c?c:"-unknown-");
 			}
-			if (info.new_level != array.level ||
-			    info.new_layout != array.layout) {
-				if (info.new_level == 5) {
-					char *c = map_num(r5layout, info.new_layout);
+			if (info->new_level != array.level ||
+			    info->new_layout != array.layout) {
+				if (info->new_level == 5) {
+					char *c = map_num(r5layout, info->new_layout);
 					printf("     New Layout : %s\n",
 					       c?c:"-unknown-");
 				}
-				if (info.new_level == 6) {
-					char *c = map_num(r6layout, info.new_layout);
+				if (info->new_level == 6) {
+					char *c = map_num(r6layout, info->new_layout);
 					printf("     New Layout : %s\n",
 					       c?c:"-unknown-");
 				}
-				if (info.new_level == 10) {
+				if (info->new_level == 10) {
 					printf("     New Layout : near=%d, %s=%d\n",
-					       info.new_layout&255,
-					       (info.new_layout&0x10000)?"offset":"far",
-					       (info.new_layout>>8)&255);
+					       info->new_layout&255,
+					       (info->new_layout&0x10000)?"offset":"far",
+					       (info->new_layout>>8)&255);
 				}
 			}
-			if (info.new_chunk != array.chunk_size)
-				printf("  New Chunksize : %dK\n", info.new_chunk/1024);
+			if (info->new_chunk != array.chunk_size)
+				printf("  New Chunksize : %dK\n", info->new_chunk/1024);
 			printf("\n");
 		} else if (e && e->percent >= 0)
 			printf("\n");
@@ -488,6 +507,7 @@ This is pretty boring
 		else
 			printf("    Number   Major   Minor   RaidDevice\n");
 	}
+	free(info);
 
 	for (d= 0; d < max_disks; d++) {
 		char *dv;
@@ -569,6 +589,7 @@ This is pretty boring
 	free(disks);
 out:
 	close(fd);
+	free(subarray);
 	return rv;
 }
 
